"""
Read BiblioSpec 2 SQLite3 spectral libraries.

For more information, see `BiblioSpec's website <https://skyline.ms/project/home/software/BiblioSpec/begin.view>`_.
"""
import sqlite3
import zlib
from dataclasses import dataclass

from typing import Iterator, List, Mapping, Tuple, Iterable, Type

import numpy as np

from pyteomics import proforma

from mzspeclib import annotation
from mzspeclib.analyte import FIRST_ANALYTE_KEY, FIRST_INTERPRETATION_KEY, Analyte
from mzspeclib.spectrum import Spectrum, SPECTRUM_NAME, CHARGE_STATE
from mzspeclib.attributes import AttributeManager, Attributed
from mzspeclib.const import (
    CHARGE_STATE,
    LIBRARY_IDENTIFIER,
    PROFORMA_ION,
    LIBRARY_NAME,
    SCAN_NUMBER,
    SOURCE_FILE,
    STRIPPED_PEPTIDE_SEQ,
    RETENTION_TIME,
    SW_VERSION,
    THEORETICAL_MASS,
    NUMBER_OF_REPLICATE_SPECTRA_AVAILABLE,
    NUMBER_OF_REPLICATE_SPECTRA_USED,
    LIBRARY_CREATION_SW,
)

from mzspeclib.backends.base import SpectralLibraryBackendBase, FORMAT_VERSION, DEFAULT_VERSION

from mzspeclib.index.base import IndexBase


class BibliospecBase:
    """Minimal base class for interacting with BiblioSpec 2 libraries"""

    connection: sqlite3.Connection

    def _correct_modifications_in_sequence(self, row: Mapping) -> proforma.ProForma:
        """
        Correct the modifications in BiblioSpec's modified peptide sequence.

        Bibliospec only stores modifications as delta masses.
        """
        mods = self.connection.execute("SELECT * FROM Modifications WHERE RefSpectraID = ?", (row['id'], )).fetchall()
        peptide = proforma.ProForma.parse(row["peptideModSeq"])
        for mod in mods:
            position = mod['position'] - 1
            mass = mod['mass']
            peptide[position][1][0] = proforma.MassModification(mass)
        return peptide




@dataclass
class BlibIndexRecord:
    """An index record read from a BiblioSpec 2 spectral library"""

    number: int
    precursor_mz: float
    precursor_charge: int
    peptide: str


class BlibIndex(BibliospecBase, IndexBase):
    """An index implementation backed by the BiblioSpec 2 SQLite3 database"""

    connection: sqlite3.Connection

    def __init__(self, connection):
        self.connection = connection

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.search(i + 1)
        elif isinstance(i, slice):
            return [self.search(j + 1) for j in range(i.start or 0, i.stop or len(self), i.step or 1)]
        else:
            raise TypeError(f"Cannot index {self.__class__.__name__} with {i}")

    def _record_from(self, row: Mapping) -> BlibIndexRecord:
        peptide_sequence = str(self._correct_modifications_in_sequence(row))
        return BlibIndexRecord(row['id'], row['precursorMZ'], row['precursorCharge'], peptide_sequence)

    def search(self, i):
        if isinstance(i, int):
            info = self.connection.execute("SELECT * FROM RefSpectra WHERE id = ?", (i, )).fetchone()
            return self._record_from(info)
        elif isinstance(i, str):
            raise NotImplementedError()

    def __iter__(self):
        return map(self._record_from, self.connection.execute("SELECT * FROM RefSpectra ORDER BY id").fetchall())

    def __len__(self):
        return self.connection.execute("SELECT count(*) FROM RefSpectra;").fetchone()[0]


class BibliospecSpectralLibrary(BibliospecBase, SpectralLibraryBackendBase):
    """Read BiblioSpec 2 SQLite3 spectral library files."""

    connection: sqlite3.Connection

    file_format = "blib"
    format_name = "bibliospec"

    @classmethod
    def has_index_preference(cls, filename) -> Type[IndexBase]:
        return BlibIndex

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self.connection = sqlite3.connect(filename)
        self.connection.row_factory = sqlite3.Row
        self.index = BlibIndex(self.connection)
        self.read_header()

    def read_header(self) -> bool:
        attribs = AttributeManager()
        attribs.add_attribute(FORMAT_VERSION, DEFAULT_VERSION)
        attribs.add_attribute(LIBRARY_CREATION_SW, "Bibliospec")

        info = self.connection.execute("SELECT * FROM LibInfo;").fetchone()
        library_id = info['libLSID']
        _, pfx_name = library_id.split("bibliospec:")
        _, name = pfx_name.split(":", 1)
        attribs.add_attribute(LIBRARY_NAME, name)
        attribs.add_attribute(LIBRARY_IDENTIFIER, library_id)
        attribs.add_attribute(SW_VERSION, f"{info['majorVersion']}.{info['minorVersion']}")
        self.attributes = attribs
        return True

    def _populate_analyte(self, analyte: Analyte, row: Mapping):
        """
        Fill an analyte with details describing a peptide sequence and inferring
        from context its traits based upon the assumptions Bibliospec makes.

        Bibliospec only stores modifications as delta masses.
        """
        peptide = self._correct_modifications_in_sequence(row)
        peptide.charge_state = row['precursorCharge']
        analyte.add_attribute(PROFORMA_ION, str(peptide))
        analyte.add_attribute(THEORETICAL_MASS, peptide.mass)
        analyte.add_attribute(STRIPPED_PEPTIDE_SEQ, row['peptideSeq'])

    def get_spectrum(self, spectrum_number: int = None, spectrum_name: str = None):
        """
        Read a spectrum from the spectrum library.

        Bibliospec does not support alternative labeling of spectra with a
        plain text name so looking up by `spectrum_name` is not supported.
        """
        if spectrum_number is None:
            raise ValueError("Only spectrum number queries are supported. spectrum_number must have an integer value")
        try:
            spectrum_number = int(spectrum_number)
        except (ValueError, TypeError):
            raise ValueError(f"spectrum_number must have an integer value, received {spectrum_number!r}") from None
        info = self.connection.execute("SELECT * FROM RefSpectra WHERE id = ?", (spectrum_number, )).fetchone()
        spectrum = self._new_spectrum()
        spectrum.key = info['id']
        spectrum.index = info['id'] - 1
        spectrum.precursor_mz = info['precursorMZ']
        spectrum.add_attribute(CHARGE_STATE, info['precursorCharge'])
        try:
            spectrum.add_attribute(RETENTION_TIME, info['retentionTime'])
        except KeyError:
            pass

        try:
            spectrum.add_attribute(NUMBER_OF_REPLICATE_SPECTRA_AVAILABLE, info['copies'])
            spectrum.add_attribute(NUMBER_OF_REPLICATE_SPECTRA_USED, 1)
        except KeyError:
            pass

        n_peaks = info['numPeaks']
        spectrum.add_attribute("MS:1003059|number of peaks", n_peaks)

        try:
            spectrum.add_attribute(
                SOURCE_FILE,
                self.connection.execute("SELECT fileName FROM SpectrumSourceFiles WHERE id = ?",
                                        (info['fileID'], )).fetchone()['fileName']
            )
        except KeyError:
            pass
        spectrum.add_attribute(
            SCAN_NUMBER,
            info["SpecIDinFile"]
        )

        analyte = self._new_analyte(1)
        self._populate_analyte(analyte, info)

        spectrum.add_analyte(analyte)

        interp = self._new_interpretation(1)
        interp.add_analyte(analyte)
        spectrum.add_interpretation(interp)

        peak_data = self.connection.execute("SELECT * FROM RefSpectraPeaks WHERE RefSpectraID = ?", (spectrum_number, )).fetchone()
        try:
            mz_array = np.frombuffer(zlib.decompress(peak_data['peakMZ']), dtype=np.float64)
        except zlib.error:
            mz_array = np.frombuffer(peak_data['peakMZ'], dtype=np.float64)
        if mz_array.size != n_peaks:
            raise ValueError(f"m/z array does not match expected peak count ({mz_array.size} != {n_peaks})")

        try:
            intensity_array = np.frombuffer(zlib.decompress(peak_data['peakIntensity']), dtype=np.float32)
        except zlib.error:
            intensity_array = np.frombuffer(peak_data['peakIntensity'], dtype=np.float32)
        if intensity_array.size != n_peaks:
            raise ValueError(f"m/z array does not match expected peak count ({intensity_array.size} != {n_peaks})")

        peak_list = []
        for i, mz in enumerate(mz_array):
            row = (mz, intensity_array[i], [], '')
            peak_list.append(row)
        spectrum.peak_list = peak_list
        return spectrum

    def read(self) -> Iterator[Spectrum]:
        for rec in self.index:
            yield self.get_spectrum(rec.number)




"""Draw an example spectrum"""

import sys
import matplotlib
matplotlib.use("agg")

from matplotlib import pyplot as plt

from mzspeclib.spectrum_library import SpectrumLibrary
from mzspeclib.draw import draw_spectrum


def main(path, spectrum_key):
    """Read spectrum `path`, select `spectrum_key` and render that spectrum to PDF"""
    lib = SpectrumLibrary(filename=path)
    spec = lib.get_spectrum(spectrum_number=spectrum_key)
    draw_spectrum(spec)
    plt.savefig(f"{path}.{spectrum_key}.annotated.pdf", bbox_inches='tight')


if __name__ == '__main__':
    try:
        path = sys.argv[1]
        index = sys.argv[2]
        main(path, int(index))
        sys.exit(0)
    except (IndexError, TypeError):
        print("USAGE: <prog> <spectrum library path> <spectrum key>")
        print("\tWrites the annotated spectrum to <spectrum library path>.<spectrum key>.annotated.pdf")
        sys.exit(1)



"""Read mzSpecLib JSON Format"""
import io
import json
import logging
import warnings

from typing import Any, Iterable, List, Dict, Mapping, Optional, Union

from pathlib import Path
from mzspeclib.cluster import SpectrumCluster

from mzspeclib.index import MemoryIndex
from mzspeclib.attributes import Attribute, AttributeManager, Attributed, AttributeSet
from mzspeclib.annotation import parse_annotation, IonAnnotationBase
from mzspeclib.analyte import Analyte, Interpretation
from mzspeclib.spectrum import Spectrum
from mzspeclib.utils import ValidationWarning
from mzspeclib.const import ATTRIBUTE_SET_NAME

from .base import (
    DEFAULT_VERSION,
    SpectralLibraryBackendBase,
    SpectralLibraryWriterBase,
    FORMAT_VERSION,
    AttributeSetTypes,
)
from .utils import open_stream


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


LIBRARY_METADATA_KEY = "attributes"
ELEMENT_ATTRIBUTES_KEY = "attributes"
SPECTRA_KEY = "spectra"
CLUSTERS_KEY = "clusters"
FORMAT_VERSION_KEY = "format_version"
ANALYTES_KEY = "analytes"
INTERPRETATIONS_KEY = "interpretations"
INTERPRETATION_MEMBERS_KEY = "members"
ID_KEY = "id"

MZ_KEY = "mzs"
INTENSITY_KEY = "intensities"
AGGREGATIONS_KEY = "aggregations"
PEAK_ANNOTATIONS_KEY = "peak_annotations"

SPECTRUM_CLASSES = "spectrum_attribute_sets"
ANALYTE_CLASSES = "analyte_attribute_sets"
INTERPRETATION_CLASSES = "interpretation_attribute_sets"
CLUSTER_CLASSES = "cluster_attribute_sets"

FORMAT_VERSION_ACC = FORMAT_VERSION.split("|")[0]


class JSONSpectralLibrary(SpectralLibraryBackendBase):
    """
    A reader for the JSON serialization of the mzSpecLib spectral library foramt.

    .. note::

        Unlike other formats readers, this type does not parse incrementally, it instead
        parses the entire JSON document in-memory and stores the parsed object structure.
        The JSON objects are then converted into :mod:`mzspeclib` types upon request. This is
        because incremental JSON parsing is substantially more difficult to do in a byte
        aware manner, not to mention slow, in Python.

        This may lead to large memory overhead when reading large libraries in JSON format.
    """

    file_format = ["mzlb.json", "mzlib.json"]
    format_name = "json"

    def __init__(self, filename, index_type=None, read_metadata=True, create_index=None):
        if index_type is None:
            index_type = MemoryIndex
        super(JSONSpectralLibrary, self).__init__(filename)
        self.buffer = {}
        self._load_buffer(self.filename)
        self.attributes = AttributeManager()
        self.index, was_initialized = index_type.from_filename(self.filename)
        if not was_initialized:
            self.create_index()
        if read_metadata:
            self.read_header()

    @classmethod
    def guess_from_filename(cls, filename: Union[str, Path, io.FileIO, Mapping]) -> bool:
        if isinstance(filename, Mapping):
            return SPECTRA_KEY in filename and LIBRARY_METADATA_KEY in filename
        return super(JSONSpectralLibrary, cls).guess_from_filename(filename)

    def _load_buffer(self, filename_or_stream: Union[str, Path, io.FileIO, Mapping]):
        if isinstance(filename_or_stream, Mapping):
            self.buffer = filename_or_stream
        else:
            if hasattr(filename_or_stream, "read"):
                self.handle = filename_or_stream
            else:
                self.handle = open_stream(filename_or_stream, "rt")
            self.buffer = json.load(self.handle)
            self.handle.close()

    def _load_attribute_sets(self, attribute_sets: dict):
        return {k: self._fill_attributes(v, AttributeSet(k, [])) for k, v in attribute_sets.items()}

    def read_header(self) -> bool:
        if self.buffer:
            self._fill_attributes(self.buffer.get(LIBRARY_METADATA_KEY), self.attributes)
            if not self.attributes.has_attribute(FORMAT_VERSION):
                warnings.warn(
                    f"Library does not have a {FORMAT_VERSION}, assuming current version",
                    category=ValidationWarning,
                )
                attributes = [Attribute(FORMAT_VERSION, DEFAULT_VERSION)] + list(self.attributes)
                self.attributes.clear()
                self.attributes._attributes_from_iterable(attributes)
            self.analyte_attribute_sets.update(self._load_attribute_sets(self.buffer.get(ANALYTE_CLASSES, {})))
            self.spectrum_attribute_sets.update(self._load_attribute_sets(self.buffer.get(SPECTRUM_CLASSES, {})))
            self.interpretation_attribute_sets.update(
                self._load_attribute_sets(self.buffer.get(INTERPRETATION_CLASSES, {}))
            )
            return True
        return False

    def create_index(self):
        for i, record in enumerate(self.buffer.get(SPECTRA_KEY, [])):
            name = None
            key = None
            for attrib in record["attributes"]:
                if attrib["accession"] == "MS:1003061":
                    name = attrib["value"]
                    if name and key:
                        break
                if attrib["accession"] == "MS:1003237":
                    key = attrib["value"]
                    if name and key:
                        break
            else:
                if not name and not key:
                    raise ValueError(f"Unidentified spectrum at index {i}")
            self.index.add(key, i, name, None, None)
        for i, record in enumerate(self.buffer.get(CLUSTERS_KEY, [])):
            key = None
            for attrib in record[ELEMENT_ATTRIBUTES_KEY]:
                if attrib["accession"] == "MS:1003267":
                    key = attrib["value"]
                    break
            else:
                if not name and not key:
                    raise ValueError(f"Unidentified spectrum cluster at index {i}")
            self.index.add_cluster(key, i, None)

    def get_spectrum(self, spectrum_number: int = None, spectrum_name: str = None) -> Spectrum:
        """
        Retrieve a single spectrum from the library.

        Parameters
        ----------
        spectrum_number : int, optional
            The index of the spectrum in the library
        spectrum_name : str, optional
            The name of the spectrum in the library

        Returns
        -------
        :class:`~.Spectrum`
        """
        if spectrum_number is not None:
            if spectrum_name is not None:
                raise ValueError("Provide only one of spectrum_number or spectrum_name")
            offset = self.index.offset_for(spectrum_number)
        elif spectrum_name is not None:
            offset = self.index.offset_for(spectrum_name)
        else:
            raise ValueError("Must provide either spectrum_number or spectrum_name argument")
        data = self.buffer[SPECTRA_KEY][offset]
        spectrum = self._make_spectrum_from_payload(data, offset)
        return spectrum

    def get_cluster(self, cluster_number: int) -> SpectrumCluster:
        offset = self.index.offset_for_cluster(cluster_number)
        data = self.buffer[CLUSTERS_KEY][offset]
        cluster = self._make_cluster_from_payload(data)
        return cluster

    def _fill_attributes(
        self, attributes: List[Dict[str, Any]], store: Attributed, context_type: AttributeSetTypes = None
    ) -> Attributed:
        last_group_id = None
        current_group_id = None
        for attrib in attributes:
            group = attrib.get("cv_param_group")
            if group is not None:
                if group != last_group_id:
                    current_group_id = store.get_next_group_identifier()
                    last_group_id = group
                    group = current_group_id
                else:
                    group = current_group_id

            if attrib["accession"] == "MS:1003212":
                store.add_attribute(ATTRIBUTE_SET_NAME, attrib["value"], group_identifier=group)
                if context_type == AttributeSetTypes.analyte:
                    self.analyte_attribute_sets[attrib["value"]].apply(store, group_identifier=group)
                elif context_type == AttributeSetTypes.spectrum:
                    self.spectrum_attribute_sets[attrib["value"]].apply(store, group_identifier=group)
                elif context_type == AttributeSetTypes.interpretation:
                    self.interpretation_attribute_sets[attrib["value"]].apply(store, group_identifier=group)
                elif context_type == AttributeSetTypes.cluster:
                    self.cluster_attribute_sets[attrib["value"]].apply(store, group_identifier=group)
                else:
                    raise ValueError(f"Could not infer which attribute set type to use for {context_type}")
            else:
                key = f'{attrib["accession"]}|{attrib["name"]}'
                if "value_accession" in attrib:
                    value = f'{attrib["value_accession"]}|{attrib["value"]}'
                else:
                    value = attrib["value"]

                store.add_attribute(key, value, group_identifier=group)
        return store

    def _make_analyte_from_payload(self, analyte_id, analyte_d: Dict) -> Analyte:
        if analyte_id != analyte_d.get("id"):
            warnings.warn(f"An analyte with explicit id {analyte_d['id']!r} does not match its key {analyte_id!r}")
        analyte = self._new_analyte(analyte_id)
        self._fill_attributes(analyte_d[ELEMENT_ATTRIBUTES_KEY], analyte, AttributeSetTypes.analyte)
        return analyte

    def _make_interpretation_from_payload(self, interpretation_id, interpretation_d: Dict) -> Interpretation:
        if interpretation_id != interpretation_d.get("id"):
            warnings.warn(
                f"An analyte with explicit id {interpretation_d['id']!r} does not match its key {interpretation_id!r}"
            )

        interpretation = self._new_interpretation(interpretation_id)
        self._fill_attributes(
            interpretation_d[ELEMENT_ATTRIBUTES_KEY], interpretation.attributes, AttributeSetTypes.interpretation
        )
        if INTERPRETATION_MEMBERS_KEY in interpretation_d:
            for member_id, member in interpretation_d[INTERPRETATION_MEMBERS_KEY].items():
                member_d = self._new_interpretation_member(member_id)
                self._fill_attributes(member[ELEMENT_ATTRIBUTES_KEY], member_d)
                interpretation.add_member_interpretation(member_d)
        return interpretation

    def _make_cluster_from_payload(self, data: Dict[str, Any]) -> SpectrumCluster:
        cluster = self._new_cluster()
        self._fill_attributes(data[ELEMENT_ATTRIBUTES_KEY], cluster, AttributeSetTypes.cluster)
        return cluster

    def _make_spectrum_from_payload(self, data: Dict, index: int = None) -> Spectrum:
        spectrum = self._new_spectrum()

        if index is not None:
            spectrum.index = index

        self._fill_attributes(data[ELEMENT_ATTRIBUTES_KEY], spectrum, AttributeSetTypes.spectrum)
        if ANALYTES_KEY in data:
            for analyte_id, analyte in data[ANALYTES_KEY].items():
                analyte_d = self._make_analyte_from_payload(analyte_id, analyte)
                spectrum.add_analyte(analyte_d)

        if INTERPRETATIONS_KEY in data:
            for interpretation_id, interpretation_d in data[INTERPRETATIONS_KEY].items():
                interpretation = self._make_interpretation_from_payload(interpretation_id, interpretation_d)
                spectrum.add_interpretation(interpretation)
                self._analyte_interpretation_link(spectrum, interpretation)
            self._default_interpretation_to_analytes(spectrum)

        peak_list = []
        n = len(data[MZ_KEY])
        mzs = data[MZ_KEY]
        intensities = data[INTENSITY_KEY]
        interpretations = data[PEAK_ANNOTATIONS_KEY]
        aggregations = data.get(AGGREGATIONS_KEY, None)
        for i in range(n):
            interpretation = interpretations[i]
            if isinstance(interpretation, str):
                interpretation = parse_annotation(interpretation)
            elif isinstance(interpretation, list):
                interpretation = [IonAnnotationBase.from_json(interp) for interp in interpretation]
            elif isinstance(interpretation, dict):
                interpretation = [IonAnnotationBase.from_json(interpretation)]
            else:
                raise TypeError(f"Cannot reconstruct interpretation from type {interpretation.__class__}")
            peak = [
                mzs[i],
                intensities[i],
                parse_annotation(interpretations[i]),
                aggregations[i] if aggregations else "",
            ]
            peak_list.append(peak)
        spectrum.peak_list = peak_list
        return spectrum

    def read(self):
        n = len(self.buffer.get(CLUSTERS_KEY, []))
        for offset in range(n):
            data = self.buffer[CLUSTERS_KEY][offset]
            cluster = self._make_cluster_from_payload(data)
            yield cluster

        n = len(self.buffer[SPECTRA_KEY])
        for offset in range(n):
            data = self.buffer[SPECTRA_KEY][offset]
            spectrum = self._make_spectrum_from_payload(data, offset)
            yield spectrum


class JSONSpectralLibraryWriter(SpectralLibraryWriterBase):
    """
    Write a spectral library to the JSON serialization of the mzSpecLib spectral library foramt.

    .. note::

        Unlike other format writers, this writer buffers the entire library in memory as JSON-compatible
        Python objects until the entire library is ready to be written out. This is because incrementally
        writing JSON is substantially more difficult to do correctly.

        This may lead to large memory overhead when writing large libraries in JSON format.
    """

    file_format = "mzlb.json"
    format_name = "json"
    default_version = "1.0"

    def __init__(self, filename, version=None, pretty_print=True, format_annotations=True, simplify=True, **kwargs):
        if version is None:
            version = self.default_version
        super(JSONSpectralLibraryWriter, self).__init__(filename)
        self._coerce_handle(self.filename)
        self.version = version
        self.pretty_print = pretty_print
        self.wrote_library = False
        self.simplify = simplify
        self.format_annotations = format_annotations
        self.buffer = {
            FORMAT_VERSION_KEY: self.version,
            LIBRARY_METADATA_KEY: [],
            SPECTRA_KEY: [],
            CLUSTERS_KEY: [],
            SPECTRUM_CLASSES: {},
            ANALYTE_CLASSES: {},
            INTERPRETATION_CLASSES: {},
        }

    def write_library(self, library: SpectralLibraryBackendBase):
        self.wrote_library = True
        return super().write_library(library)

    def _split_compound_value(self, value):
        value = str(value)
        # Don't process quoted values
        if value.startswith('"'):
            return [value]
        components = value.split("|", 1)
        return components

    def write_header(self, library: SpectralLibraryBackendBase):
        """Write the library header and other global metadata"""
        attributes = self._format_attributes(library.attributes)
        self.buffer[LIBRARY_METADATA_KEY] = attributes
        self.buffer[SPECTRUM_CLASSES] = {
            c.name: self._format_attributes(c.attributes) for c in library.spectrum_attribute_sets.values()
        }
        self.buffer[ANALYTE_CLASSES] = {
            c.name: self._format_attributes(c.attributes) for c in library.analyte_attribute_sets.values()
        }
        self.buffer[INTERPRETATION_CLASSES] = {
            c.name: self._format_attributes(c.attributes) for c in library.interpretation_attribute_sets.values()
        }

    def _format_attributes(self, attributes_manager: Attributed, attribute_sets: Optional[List[str]] = None) -> List:
        if attribute_sets is None:
            attribute_sets = []

        attributes = []
        for attribute in attributes_manager:
            if attribute.owner_id and attribute.owner_id in attribute_sets or attribute.owner_id == 'all':
                continue
            reformed_attribute = {}
            key = attribute.key
            value = attribute.value
            if attribute.group_id is not None:
                cv_param_group = attribute.group_id
                reformed_attribute["cv_param_group"] = cv_param_group

            term = None
            components = key.split("|", 1)
            if len(components) == 2:
                accession, name = components
                reformed_attribute["accession"] = accession
                reformed_attribute["name"] = name
                try:
                    term = self.find_term_for(accession)
                except KeyError:
                    pass
            else:
                raise ValueError(f"Unsupported number of items in components: {components}")

            components = self._split_compound_value(value)
            if len(components) == 2:
                value_accession, value = components
                reformed_attribute["value_accession"] = value_accession
                reformed_attribute["value"] = value
            elif len(components) == 1:
                # If an attribute could take on a JSON-incompatible type, we need to
                # cast it prior to writing it out.
                if not isinstance(value, (str, int, float, list)):
                    if term is not None:
                        value = term.value_type.format(value)
                    else:
                        value = str(value)
                reformed_attribute["value"] = value
            else:
                raise ValueError(f"Unsupported number of items in components: {components}")
            attributes.append(reformed_attribute)
        return attributes

    def write_cluster(self, cluster: SpectrumCluster):
        attributes = self._format_attributes(cluster.attributes)
        payload = {ELEMENT_ATTRIBUTES_KEY: attributes}
        self.buffer[CLUSTERS_KEY].append(payload)

    def write_spectrum(self, spectrum: Spectrum):
        mzs = []
        intensities = []
        annotations = []
        aggregations = []
        for peak in spectrum.peak_list:
            mzs.append(peak[0])
            intensities.append(peak[1])
            if self.format_annotations:
                annotations.append("?" if not peak[2] else ",".join(map(str, peak[2])))
            else:
                annotations.append([c.to_json() for c in peak[2]])
            aggregations.append(peak[3])

        #### Organize the attributes from the simple list into the appropriate JSON format
        attributes = self._format_attributes(
            list(self._filter_attributes(spectrum.attributes, self._not_entry_index)), spectrum.attribute_sets
        )

        analytes = {}
        for analyte in spectrum.analytes.values():
            analyte_d = {
                ID_KEY: analyte.id,
                ELEMENT_ATTRIBUTES_KEY: self._format_attributes(analyte, analyte.attribute_sets),
            }
            analytes[analyte.id] = analyte_d

        interpretations = {}
        for interpretation in spectrum.interpretations.values():
            if len(analytes) == 1:
                attribs_of = self._filter_attributes(interpretation, self._not_analyte_mixture_term)
            else:
                attribs_of = interpretation.attributes
            interpretation_d = {
                ID_KEY: interpretation.id,
                ELEMENT_ATTRIBUTES_KEY: self._format_attributes(attribs_of, interpretation.attribute_sets),
            }
            interpretations[interpretation.id] = interpretation_d
            if interpretation.member_interpretations:
                members_d = interpretation_d[INTERPRETATION_MEMBERS_KEY] = {}
                for member in interpretation.member_interpretations.values():
                    members_d[member.id] = {
                        ID_KEY: member.id,
                        ELEMENT_ATTRIBUTES_KEY: self._format_attributes(member, member.attribute_sets),
                    }

        payload = {
            ELEMENT_ATTRIBUTES_KEY: attributes,
            MZ_KEY: mzs,
            INTENSITY_KEY: intensities,
            PEAK_ANNOTATIONS_KEY: annotations,
            AGGREGATIONS_KEY: aggregations,
            ANALYTES_KEY: analytes,
            INTERPRETATIONS_KEY: interpretations,
        }
        if not any(aggregations):
            payload.pop(AGGREGATIONS_KEY)

        self.buffer[SPECTRA_KEY].append(payload)

    def _flush(self):
        # If we know we're writing a complete library, skip the probably-doing-too-many-things
        # formatting logic for single vs. many spectra.
        if self.wrote_library:
            if self.pretty_print:
                json.dump(self.buffer, self.handle, indent=2, sort_keys=True)
            else:
                json.dump(self.buffer, self.handle)
        else:
            # We don't have a header section to format, so write just the spectra,
            # and if the number of spectra is one and the simplify flag is true,
            # skip the wrapping array
            spectra = self.buffer[SPECTRA_KEY]
            n_spectra = len(spectra)
            if n_spectra == 1 and self.simplify:
                if self.pretty_print:
                    json.dump(spectra[0], self.handle, indent=2, sort_keys=True)
                else:
                    json.dump(spectra[0], self.handle)
            else:
                if self.pretty_print:
                    json.dump(spectra, self.handle, indent=2, sort_keys=True)
                else:
                    json.dump(spectra, self.handle)

    def close(self):
        if not self.handle.closed:
            self._flush()
            self.handle.close()


def format_spectrum(spectrum: Spectrum, pretty_print=True, **kwargs) -> str:
    buffer = io.StringIO()
    with JSONSpectralLibraryWriter(buffer, pretty_print=pretty_print, **kwargs) as writer:
        writer.write_spectrum(spectrum)
        writer.flush()
        return buffer.getvalue()



#!/usr/bin/env python3
# pragma: no cover
from __future__ import print_function
import sys
def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

import os
from datetime import datetime

from sqlalchemy import Column, ForeignKey, Integer, Float, String, DateTime, Text, PickleType, LargeBinary
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy import desc
from sqlalchemy import inspect

Base = declarative_base()

debug = False

#### Define the database tables as classes
class LibraryRecord(Base):
  __tablename__ = 'library_record'
  library_record_id = Column(Integer, primary_key=True)
  id_name = Column(String(255), nullable=False)
  version = Column(String(255), nullable=False)
  status = Column(String(255), nullable=False)
  original_name = Column(String(255), nullable=False)
  original_checksum = Column(String(255), nullable=True)
  n_entries = Column(Integer, nullable=True)
  record_datetime = Column(DateTime, nullable=False)


class SpectrumLibraryCollection:
    """
    SpectrumLibraryCollection - Class for a collection of spectrum libraries

    Attributes
    ----------
    filename : string
        Filename of the SQLite database file that contains information about the collection of libraries available.

    Methods
    -------
    create - Create a new spectral library collection
    show - Return a string that summarizes the state of the collection
    get_libraries - Return a list of available libraries
    get_library - Return attributes of a specific library
    add_library - Add a new library
    create_index - Create a master index from all the constituent library indexes to be able to find spectra in any library
    find_spectra - Return a list of spectra given query constraints

    """


    def __init__(self, filename=None):
        """
        __init__ - SpectrumLibraryCollection constructor

        Parameters
        ----------
        filename : string
            Filename of the SQLite database file that contains information about the collection of libraries available.

        """

        self.filename = filename
        if os.path.exists(self.filename):
            if debug: eprint(f"INFO: Library collection {self.filename} exists")
            self.connect()
        else:
            if debug: eprint(f"INFO: Library collection {self.filename} not found. Will create new.")
            self.createDatabase()


    #### Destructor
    def __del__(self):
        self.disconnect()


    #### Define getter/setter for attribute filename
    @property
    def filename(self):
        return(self._filename)
    @filename.setter
    def filename(self, filename):
        self._filename = filename


    #### Define attribute session
    @property
    def session(self) -> str:
        return self._session

    @session.setter
    def session(self, session: str):
        self._session = session


    #### Define attribute engine
    @property
    def engine(self) -> str:
        return self._engine

    @engine.setter
    def engine(self, engine: str):
        self._engine = engine


    #### Delete and create the database. Careful!
    def createDatabase(self):
        if os.path.exists(self.filename):
            eprint("INFO: Deleting previous database file " + self.filename)
            os.remove(self.filename)
        if debug: eprint("INFO: Creating database " + self.filename)
        engine = create_engine("sqlite:///"+self.filename)
        Base.metadata.create_all(engine)
        self.connect()


    #### Create and store a database connection
    def connect(self):
        if debug: eprint("INFO: Connecting to database " + self.filename)
        engine = create_engine("sqlite:///"+self.filename)
        DBSession = sessionmaker(bind=engine)
        session = DBSession()
        self.session = session
        self.engine = engine


    #### Destroy the database connection
    def disconnect(self):
        if debug: eprint("INFO: Disconnecting from database " + self.filename)
        session = self.session
        engine = self.engine
        session.close()
        engine.dispose()


    def create(self, overwrite_existing=False):
        """
        create - Create a new spectral library collection

        Extended description of function.

        Parameters
        ----------
        overwrite_existing : boolean
            Set to true in order to write over the previous file if it exists

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here
        if debug: eprint("INFO: Creating database " + self.filename)
        if os.path.exists(self.filename):
            os.remove(self.filename)
        engine = create_engine("sqlite:///"+self.filename)
        Base.metadata.create_all(engine)
        self.connect()
        return()


    def get_libraries(self):
        """
        get_libraries - Return a list of available libraries

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here
        if debug: eprint("INFO: Fetching all libraries")
        session = self.session
        libraries = session.query(LibraryRecord).all()
        return(libraries)


    def get_library(self,identifier=None,version=None,filename=None):
        """
        get_library - Return attributes of a specific library

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        session = self.session
        if identifier is not None and identifier > "":
            libraries = session.query(LibraryRecord).filter(LibraryRecord.id_name==identifier).order_by(desc(LibraryRecord.version)).all()
            if len(libraries) == 0:
                raise Exception(f"No library with identifier {identifier} found")
            else:
                libraryListStr = ""
                for library in libraries:
                    libraryListStr += library.version+","
                    if version is not None and version > "":
                        if version == library.version:
                            return(library)
                if version is not None and version > "":
                    raise Exception(f"Unable to find version {version} of library {identifier}")
            if len(libraries) == 1:
                return(library)
            raise Exception(f"There are several version of this library ({libraryListStr}). Please specify a version")
            return()

        elif filename is not None and filename > "":
            raise Exception("Search by filename not implemented")
        else:
            raise Exception("Not enough information to find library")

        return()


    def add_library(self, original_name, version="1"):
        """
        add_library - Add a new library

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here
        if debug: eprint("INFO: Adding a library entry")
        session = self.session
        library_record = LibraryRecord(id_name="PXL000000", version=version,
            status="initial_add",original_name=original_name,record_datetime=datetime.now())
        session.add(library_record)
        session.flush()
        assert(library_record.library_record_id)
        idstr = str(library_record.library_record_id)
        if debug: eprint(f"INFO: Returned id={idstr}")
        idstr_length = len(idstr)
        assert(idstr_length)
        padding = "000000"
        new_idstr = "PXL" + padding[0:len(padding)-idstr_length] + idstr
        library_record.id_name = new_idstr
        session.flush()
        session.commit()
        return()


    def update_library_metadata(self, id, version=None):
        """
        update_library_metadata - Update the metadata associated with a library

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here
        if debug: eprint(f"INFO: Updating the library entry with id {id}")
        session = self.session
        library = session.query(LibraryRecord).filter(LibraryRecord.library_record_id==id).first()
        if library is not None:
            if version is not None:
                print(f"    Updating version to {version}")
                library.version = version
            session.flush()
            session.commit()
        else:
            print(f"ERROR: Library entry with id {id} not found")


    def create_index(self):
        """
        create_index - Create a master index from all the constituent library indexes to be able to find spectra in any library

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here

        return()


    def find_spectra(self):
        """
        find_spectra - Return a list of spectra given query constraints

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here

        return()


#### Example using this class
def example():

    #### Create a new or attach to existing library collection
    spec_lib_collection = SpectrumLibraryCollection("zzTest.sqlite")

    #### Add a library
    spec_lib_collection.add_library()

    #### Show all libraries
    libraries = spec_lib_collection.get_libraries()
    if ( libraries is None ):
        print("The library collection is empty")
    else:
        for library in libraries:
            print(library.library_record_id,library.id_name,library.original_name)

    return()


#### Example using this class
def example2():

    #### Create a new or attach to existing library collection
    spec_lib_collection = SpectrumLibraryCollection("../spectralLibraries/SpectrumLibraryCollection.sqlite")

    try:
        library = spec_lib_collection.get_library(identifier="PXL000003", version="05-24-2011")
    except Exception as error:
        print("ERROR:",error)
        return()
    print("\t".join([str(library.library_record_id),library.id_name,library.version,library.original_name]))
    return()


#### If this class is run from the command line, perform a short little test to see if it is working correctly
def main():

    #### Run an example
    example2()
    return()


if __name__ == "__main__": main()

